package gov.anl.aps.dm.portal.controllers;

import gov.anl.aps.dm.common.exceptions.DmException;
import gov.anl.aps.dm.common.exceptions.InvalidRequest;
import gov.anl.aps.dm.portal.model.beans.DmEntityDbFacade;
import gov.anl.aps.dm.portal.model.entities.DmEntity;
import gov.anl.aps.dm.common.utilities.CollectionUtility;
import gov.anl.aps.dm.common.utilities.StringUtility;
import gov.anl.aps.dm.portal.utilities.SessionUtility;

import java.io.Serializable;

import java.util.List;
import javax.annotation.PostConstruct;
import javax.faces.model.DataModel;
import javax.faces.model.ListDataModel;
import javax.faces.model.SelectItem;
import org.apache.log4j.Logger;
import org.primefaces.event.data.FilterEvent;

public abstract class DmEntityController<EntityType extends DmEntity, FacadeType extends DmEntityDbFacade<EntityType>> implements Serializable {

    private static final Logger logger = Logger.getLogger(DmEntityController.class.getName());
    private static final int NUMBER_OF_ROWS_FOR_DISPLAY_LIST = 25;
    private static final int NUMBER_OF_ROWS_FOR_SELECT_LIST = 10;

    protected EntityType current = null;

    private DataModel listDataModel = null;
    private List<EntityType> filteredObjectList = null;

    private DataModel selectDataModel = null;
    private List<EntityType> selectedObjectList = null;

    private int displayNumberOfItemsPerPage = NUMBER_OF_ROWS_FOR_DISPLAY_LIST;
    private int selectNumberOfItemsPerPage = NUMBER_OF_ROWS_FOR_SELECT_LIST;

    protected String breadcrumbViewParam = null;
    protected String breadcrumbObjectIdViewParam = null;
    
    public DmEntityController() {
    }

    @PostConstruct
    public void initialize() {
    }

    /**
     * Navigate to invalid request error page.
     *
     * @param error error message
     */
    public void handleInvalidSessionRequest(String error) {
        SessionUtility.setLastSessionError(error);
        SessionUtility.navigateTo("/views/error/invalidRequest?faces-redirect=true");
    }
    
    protected abstract FacadeType getFacade();

    protected abstract EntityType createEntityInstance();

    public abstract String getEntityTypeName();

    public String getDisplayEntityTypeName() {
        return getEntityTypeName();
    }

    public abstract String getCurrentEntityInstanceName();

    public EntityType getCurrent() {
        return current;
    }

    public void setCurrent(EntityType current) {
        this.current = current;
    }

    /**
     * Find entity instance by id.
     *
     * @param id entity instance id
     * @return entity instance
     */
    public EntityType findById(Integer id) {
        return null;
    }
    
    public void selectByRequestParams() {
    }

    public EntityType getSelected() {
        if (current == null) {
            current = createEntityInstance();
        }
        return current;
    }

    public boolean isEntitySelected() {
        return (current != null);
    }

    public void onFilterChange(FilterEvent filterEvent) {
    }

    /**
     * Clear entity list filters.
     *
     * This method should be overridden in any derived controller class that has
     * its own filters.
     */
    public void clearListFilters() {
    }

    /**
     * Clear entity selection list filters.
     *
     * This method should be overridden in any derived controller class that has
     * its own select filters.
     */
    public void clearSelectFilters() {
    }

    public String resetList() {
        logger.debug("Resetting list data model for " + getDisplayEntityTypeName());
        clearListFilters();
        resetListDataModel();
        return prepareList();
    }

    public String prepareList() {
        logger.debug("Preparing list data model for " + getDisplayEntityTypeName());
        current = null;
        return "list?faces-redirect=true";
    }

    /**
     * Process view request parameters.
     * 
     * If request is not valid, user will be redirected to appropriate error
     * page.
     */
    public void processViewRequestParams() {
        try {
            EntityType entity = selectByViewRequestParams();
            if (entity != null) {
                prepareEntityView(entity);
            }
        } catch (DmException ex) {
            handleInvalidSessionRequest(ex.getErrorMessage());
        }
    }

    /**
     * Set breadcrumb variables from request parameters.
     */
    protected void setBreadcrumbRequestParams() {
        if (breadcrumbViewParam == null) {
            breadcrumbViewParam = SessionUtility.getRequestParameterValue("breadcrumb");
        }
        if (breadcrumbObjectIdViewParam == null) {
            breadcrumbObjectIdViewParam = SessionUtility.getRequestParameterValue("breadcrumbObjectId");
        }
    }

    /**
     * Select current entity instance for view from request parameters.
     * 
     * @return selected entity instance
     * @throws DmException in case of invalid request parameter values
     */
    public EntityType selectByViewRequestParams() throws DmException {
        setBreadcrumbRequestParams();
        Integer idParam = null;
        String paramValue = SessionUtility.getRequestParameterValue("id");
        try {
            if (paramValue != null) {
                idParam = Integer.parseInt(paramValue);
            }
        } catch (NumberFormatException ex) {
            throw new InvalidRequest("Invalid value supplied for " + getDisplayEntityTypeName() + " id: " + paramValue);
        }
        if (idParam != null) {
            EntityType entity = findById(idParam);
            if (entity == null) {
                throw new InvalidRequest(StringUtility.capitalize(getDisplayEntityTypeName()) + " id " + idParam + " does not exist.");
            }
            setCurrent(entity);
            return entity;
        } else if (current == null || current.getId() == null) {
            throw new InvalidRequest(StringUtility.capitalize(getDisplayEntityTypeName()) + " has not been selected.");
        }
        return current;
    }

    /**
     * Follow breadcrumb if it is set, or prepare entity list view.
     * 
     * @return previous view if breadcrumb parameters are set, or entity list
     * view otherwise
     */
    public String followBreadcrumbOrPrepareList() {
        String loadView = breadcrumbViewParam;
        if (loadView == null) {
            loadView = prepareList();
        } else {
            if (breadcrumbObjectIdViewParam != null) {
                Integer entityId = Integer.parseInt(breadcrumbObjectIdViewParam);
                loadView = breadcrumbViewParam + "?faces-redirect=true&id=" + entityId;
            }
        }
        breadcrumbViewParam = null;
        breadcrumbObjectIdViewParam = null;
        return loadView;
    }

    public String prepareView(EntityType entity) {
        logger.debug("Preparing view for " + entity.toString());
        current = entity;
        prepareEntityView(entity);
        return view();
    }

    protected void prepareEntityView(EntityType entity) {
    }

    public void clear() {
    }

    public String view() {
        return "view?faces-redirect=true";
    }

    public String prepareCreate() {
        current = createEntityInstance();
        return "create?faces-redirect=true";
    }

    protected void prepareEntityInsert(EntityType entity) throws DmException {
    }

    public String create() {
        try {
            EntityType newEntity = current;
            prepareEntityInsert(current);
            getFacade().create(current);
            SessionUtility.addInfoMessage("Success", "Created " + getDisplayEntityTypeName() + " " + getCurrentEntityInstanceName() + ".");
            resetListDataModel();
            current = newEntity;
            return prepareList();
        } catch (DmException | RuntimeException ex) {
            SessionUtility.addErrorMessage("Error", "Could not create " + getDisplayEntityTypeName() + ": " + ex.getMessage());
            return null;
        }
    }

    public String prepareEdit(EntityType entity) {
        current = entity;
        return edit();
    }

    public String edit() {
        resetSelectDataModel();
        return "edit?faces-redirect=true";
    }

    protected void prepareEntityUpdate(EntityType entity) throws DmException {
    }

    public String update() {
        try {
            logger.debug("Updating " + getDisplayEntityTypeName() + " " + getCurrentEntityInstanceName());
            EntityType updatedEntity = current;
            prepareEntityUpdate(updatedEntity);
            getFacade().edit(updatedEntity);
            SessionUtility.addInfoMessage("Success", "Updated " + getDisplayEntityTypeName() + " " + getCurrentEntityInstanceName() + ".");
            resetListDataModel();
            current = updatedEntity;
            return view();
        } catch (DmException ex) {
            SessionUtility.addErrorMessage("Error", "Could not update " + getDisplayEntityTypeName() + ": " + ex.getMessage());
            return null;
        } catch (RuntimeException ex) {
            SessionUtility.addErrorMessage("Error", "Could not update " + getDisplayEntityTypeName() + ": " + getObjectAlreadyExistMessage(current));
            return null;
        }
    }

    protected String getObjectAlreadyExistMessage(EntityType entity) {
        return "";
    }

    protected void prepareEntityUpdateOnRemoval(EntityType entity) throws DmException {
    }
    
    public String updateOnRemoval() {
        try {
            logger.debug("Updating " + getDisplayEntityTypeName() + " " + getCurrentEntityInstanceName());
            prepareEntityUpdateOnRemoval(current);
            getFacade().edit(current);
            EntityType updatedEntity = current;
            SessionUtility.addInfoMessage("Success", "Updated " + getDisplayEntityTypeName() + " " + getCurrentEntityInstanceName() + ".");
            resetListDataModel();
            resetSelectDataModel();
            current = updatedEntity;
            return view();
        } catch (DmException ex) {
            SessionUtility.addErrorMessage("Error", "Could not update " + getDisplayEntityTypeName() + ": " + ex.getMessage());
            return null;
        } catch (RuntimeException ex) {
            logger.error("Could not update " + getDisplayEntityTypeName() + " "
                    + getCurrentEntityInstanceName() + ": " + ex.getMessage());
            SessionUtility.addErrorMessage("Error", "Could not update " + getDisplayEntityTypeName() + ": " + ex.getMessage());
            return null;
        }
    }

    protected void prepareEntityDestroy(EntityType entity) throws DmException {
    }

    public void destroy(EntityType entity) {
        current = entity;
        destroy();
    }

    public String destroy() {
        if (current == null) {
            logger.warn("Current item is not set");
            // Do nothing if current item is not set.
            return null;
        }
        try {
            logger.debug("Destroying " + getCurrentEntityInstanceName());
            getFacade().remove(current);
            SessionUtility.addInfoMessage("Success", "Deleted " + getDisplayEntityTypeName() + " " + getCurrentEntityInstanceName() + ".");
            resetListDataModel();
            return prepareList();
        } catch (Exception ex) {
            SessionUtility.addErrorMessage("Error", "Could not delete " + getDisplayEntityTypeName() + ": " + ex.getMessage());
            return null;
        }
    }

    public DataModel createListDataModel() {
        return new ListDataModel(getFacade().findAll());
    }

    public DataModel getListDataModel() {
        if (listDataModel == null) {
            listDataModel = createListDataModel();
        }
        return listDataModel;
    }

    public void prepareEntityListForSelection(List<EntityType> selectEntityList) {
    }

    public DataModel createSelectDataModel() {
        List<EntityType> selectEntityList = getFacade().findAll();
        prepareEntityListForSelection(selectEntityList);
        return new ListDataModel(selectEntityList);
    }

    public DataModel getSelectDataModel() {
        if (selectDataModel == null) {
            selectDataModel = createSelectDataModel();
        }
        return selectDataModel;
    }

    public DataModel getItems() {
        return getListDataModel();
    }

    public List<EntityType> getSelectedObjectListAndResetSelectDataModel() {
        List<EntityType> returnList = selectedObjectList;
        resetSelectDataModel();
        return returnList;
    }

    public List<EntityType> getSelectedObjectList() {
        return selectedObjectList;
    }

    public List<EntityType> getFilteredObjectList() {
        return filteredObjectList;
    }

    public List<EntityType> getFilteredItems() {
        return filteredObjectList;
    }

    public void resetSelectedObjectList() {
        selectedObjectList = null;
    }

    public void setSelectedObjectList(List<EntityType> selectedObjectList) {
        this.selectedObjectList = selectedObjectList;
    }

    public void setFilteredObjectList(List<EntityType> filteredObjectList) {
        this.filteredObjectList = filteredObjectList;
    }

    public void setFilteredItems(List<EntityType> filteredItems) {
        this.filteredObjectList = filteredItems;
    }

    public void resetListDataModel() {
        listDataModel = null;
        filteredObjectList = null;
        current = null;
    }

    public void resetSelectDataModel() {
        selectDataModel = null;
        selectedObjectList = null;
    }

    public List<EntityType> getAvailableItems() {
        return getFacade().findAll();
    }

    public EntityType getEntity(Integer id) {
        return getFacade().find(id);
    }

    public SelectItem[] getAvailableItemsForSelectMany() {
        return CollectionUtility.getSelectItems(getFacade().findAll(), false);
    }

    public SelectItem[] getAvailableItemsForSelectOne() {
        return CollectionUtility.getSelectItems(getFacade().findAll(), true);
    }

    public String getCurrentViewId() {
        return SessionUtility.getCurrentViewId();
    }

    public static String displayEntityList(List<?> entityList) {
        String itemDelimiter = ", ";
        return CollectionUtility.displayItemListWithoutOutsideDelimiters(entityList, itemDelimiter);
    }

    public int getDisplayNumberOfItemsPerPage() {
        return displayNumberOfItemsPerPage;
    }

    public void setDisplayNumberOfItemsPerPage(int displayNumberOfItemsPerPage) {
        this.displayNumberOfItemsPerPage = displayNumberOfItemsPerPage;
    }

    public int getSelectNumberOfItemsPerPage() {
        return selectNumberOfItemsPerPage;
    }

    public void setSelectNumberOfItemsPerPage(int selectNumberOfItemsPerPage) {
        this.selectNumberOfItemsPerPage = selectNumberOfItemsPerPage;
    }

}
