package gov.anl.aps.dm.portal.controllers;

import gov.anl.aps.dm.api.ExperimentDsApi;
import gov.anl.aps.dm.common.exceptions.InvalidRequest;
import gov.anl.aps.dm.common.exceptions.DmException;
import gov.anl.aps.dm.common.exceptions.ObjectAlreadyExists;
import gov.anl.aps.dm.portal.model.entities.UserInfo;
import gov.anl.aps.dm.portal.model.beans.UserInfoDbFacade;
import gov.anl.aps.dm.common.utilities.CryptUtility;
import gov.anl.aps.dm.portal.model.beans.UserExperimentRoleDbFacade;
import gov.anl.aps.dm.portal.model.beans.ExperimentRoleTypeDbFacade;
import gov.anl.aps.dm.portal.model.entities.Experiment;
import gov.anl.aps.dm.portal.model.entities.ExperimentRoleType;
import gov.anl.aps.dm.portal.model.entities.UserExperimentRole;
import gov.anl.aps.dm.portal.model.entities.UserExperimentRolePK;
import gov.anl.aps.dm.portal.utilities.DmApiFactory;
import gov.anl.aps.dm.portal.utilities.SessionUtility;

import java.util.List;
import javax.ejb.EJB;
import javax.enterprise.context.SessionScoped;
import javax.faces.component.UIComponent;
import javax.faces.context.FacesContext;
import javax.faces.convert.Converter;
import javax.faces.convert.FacesConverter;
import javax.inject.Named;
import org.apache.log4j.Logger;

@Named("userInfoController")
@SessionScoped
public class UserInfoController extends DmEntityController<UserInfo, UserInfoDbFacade> {

    private static final Logger logger = Logger.getLogger(UserInfoController.class.getName());

    @EJB
    private UserInfoDbFacade userInfoDbFacade;
    @EJB
    private UserExperimentRoleDbFacade userExperimentRoleDbFacade;
    @EJB
    private ExperimentRoleTypeDbFacade experimentRoleTypeDbFacade;

    private String passwordEntry = null;
    private UserExperimentRole currentUserExperimentRole = null;

    private String filterByBadge = null;
    private String filterByUsername = null;
    private String filterByLastName = null;
    private String filterByFirstName = null;
    private String filterByEmail = null;
    private String filterByGlobusUsername = null;
    private String filterByDescription = null;

    private String selectFilterByBadge = null;
    private String selectFilterByUsername = null;
    private String selectFilterByLastName = null;
    private String selectFilterByFirstName = null;
    private String selectFilterByEmail = null;
    private String selectFilterByGlobusUsername = null;
    private String selectFilterByDescription = null;

    public UserInfoController() {
    }

    @Override
    protected UserInfoDbFacade getFacade() {
        return userInfoDbFacade;
    }

    @Override
    protected UserInfo createEntityInstance() {
        return new UserInfo();
    }

    @Override
    public String getEntityTypeName() {
        return "user";
    }

    @Override
    public String getCurrentEntityInstanceName() {
        if (getCurrent() != null) {
            return getCurrent().getUsername();
        }
        return "";
    }

    @Override
    public UserInfo findById(Integer id) {
        return userInfoDbFacade.findById(id);
    }

    @Override
    public List<UserInfo> getAvailableItems() {
        return super.getAvailableItems();
    }

    @Override
    public String prepareEdit(UserInfo user) {
        passwordEntry = null;
        return super.prepareEdit(user);
    }

    @Override
    public void prepareEntityInsert(UserInfo user) throws DmException {
        checkMandatoryParameters(user);
        UserInfo existingUser = userInfoDbFacade.findByUsername(user.getUsername());
        if (existingUser != null) {
            throw new ObjectAlreadyExists("User " + user.getUsername() + " already exists.");
        }
        if (passwordEntry != null && !passwordEntry.isEmpty()) {
            logger.debug("Will crypt provided password for user " + user.getUsername());
            String cryptedPassword = CryptUtility.cryptPasswordWithPbkdf2(passwordEntry);
            user.setPassword(cryptedPassword);
        }
        passwordEntry = null;
        user.setIsLocalUser(true);
        logger.debug("Inserting new user " + user.getUsername());
    }

    private void checkMandatoryParameters(UserInfo user) throws InvalidRequest {
        if ((user.getUsername() == null) || (user.getUsername().length() == 0)) {
            throw new InvalidRequest("Missing username.");
        }
        if ((user.getLastName() == null) || (user.getLastName().length() == 0)) {
            throw new InvalidRequest("Missing user last name.");
        }
        if ((user.getFirstName() == null) || (user.getFirstName().length() == 0)) {
            throw new InvalidRequest("Missing user first name.");
        }
    }

    @Override
    public void prepareEntityUpdate(UserInfo user) throws DmException {
        checkMandatoryParameters(user);
        if (passwordEntry != null && !passwordEntry.isEmpty()) {
            logger.debug("Will crypt provided password for user " + user.getUsername());
            String cryptedPassword = CryptUtility.cryptPasswordWithPbkdf2(passwordEntry);
            user.setPassword(cryptedPassword);
        }
        passwordEntry = null;
    }

    @Override
    protected String getObjectAlreadyExistMessage(UserInfo user) {
        if (user == null) {
            return null;
        }
        return "User " + user.getUsername() + " already exists.";
    }

    public String prepareSessionUserEdit(String viewPath) {
        UserInfo sessionUser = (UserInfo) SessionUtility.getUser();
        if (sessionUser == null) {
            return null;
        }
        prepareEdit(sessionUser);
        return viewPath + "?faces-redirect=true";
    }

    public String getPasswordEntry() {
        return passwordEntry;
    }

    public void setPasswordEntry(String passwordEntry) {
        this.passwordEntry = passwordEntry;
    }

    public UserExperimentRole getCurrentUserExperimentRole() {
        if (currentUserExperimentRole == null) {
            currentUserExperimentRole = new UserExperimentRole();
        }
        return currentUserExperimentRole;
    }

    public void setCurrentUserExperimentRole(UserExperimentRole currentUserExperimentRole) {
        this.currentUserExperimentRole = currentUserExperimentRole;
    }

    public void prepareAddUserExperimentRole(UserInfo user) {
        UserExperimentRole userExperimentRole = getCurrentUserExperimentRole();
        userExperimentRole.setUserInfo(user);
    }

    private void addUserExperimentRole() {
        UserInfo user = currentUserExperimentRole.getUserInfo();
        Experiment experiment = currentUserExperimentRole.getExperiment();
        ExperimentRoleType roleType = currentUserExperimentRole.getExperimentRoleType();
        UserExperimentRolePK userExperimentRolePK = new UserExperimentRolePK();
        userExperimentRolePK.setExperimentId(experiment.getId());
        userExperimentRolePK.setUserId(user.getId());
        userExperimentRolePK.setRoleTypeId(roleType.getId());
        currentUserExperimentRole.setUserExperimentRolePK(userExperimentRolePK);

        UserExperimentRole existingRole = userExperimentRoleDbFacade.findByUserAndExperiment(user.getId(),
                experiment.getId());
        if (existingRole != null) {
            SessionUtility.addErrorMessage("Error", "Could not add user experiment role: User "
                    + user.getUsername()
                    + " is already associated with experiment "
                    + existingRole.getExperiment().getName() + " as "
                    + existingRole.getExperimentRoleType().getName() + ".");
            return;
        }
        List<UserExperimentRole> userExperimentRoleList = user.getUserExperimentRoleList();
        userExperimentRoleList.add(0, currentUserExperimentRole);
        update();
        notifyUpdateExperiment(experiment);
        currentUserExperimentRole = null;
    }

    public void addUserExperimentRolePi() {
        currentUserExperimentRole.setExperimentRoleType(experimentRoleTypeDbFacade.findPiRoleType());
        addUserExperimentRole();
    }

    public void addUserExperimentRoleUser() {
        currentUserExperimentRole.setExperimentRoleType(experimentRoleTypeDbFacade.findUserRoleType());
        addUserExperimentRole();
    }

    public void deleteUserExperimentRole(UserExperimentRole userExperimentRole) {
        UserInfo user = getCurrent();
        userExperimentRoleDbFacade.remove(userExperimentRole);
        List<UserExperimentRole> userExperimentRoleList = user.getUserExperimentRoleList();
        userExperimentRoleList.remove(userExperimentRole);
        updateOnRemoval();
        notifyUpdateExperiment(userExperimentRole.getExperiment());
    }

    public void notifyUpdateExperiment(Experiment experiment) {
        // Notify DS Web Service
        try {
            ExperimentDsApi api = DmApiFactory.getExperimentDsApi();
            api.updateExperiment(experiment.getName());
        } catch (DmException ex) {
            logger.error("Could not notify Data Storage Service: " + ex);
            SessionUtility.addErrorMessage("Error", "Unable to notify Data Storage Service: " + ex.getErrorMessage());
        }
    }

    @Override
    public void clearListFilters() {
        filterByBadge = null;
        filterByUsername = null;
        filterByLastName = null;
        filterByFirstName = null;
        filterByEmail = null;
        filterByGlobusUsername = null;
        filterByDescription = null;
    }

    @Override
    public void clearSelectFilters() {
        selectFilterByBadge = null;
        selectFilterByUsername = null;
        selectFilterByLastName = null;
        selectFilterByFirstName = null;
        selectFilterByEmail = null;
        selectFilterByGlobusUsername = null;
        selectFilterByDescription = null;
    }

    public String getFilterByBadge() {
        return filterByBadge;
    }

    public void setFilterByBadge(String filterByBadge) {
        this.filterByBadge = filterByBadge;
    }

    public String getFilterByUsername() {
        return filterByUsername;
    }

    public void setFilterByUsername(String filterByUsername) {
        this.filterByUsername = filterByUsername;
    }

    public String getFilterByLastName() {
        return filterByLastName;
    }

    public void setFilterByLastName(String filterByLastName) {
        this.filterByLastName = filterByLastName;
    }

    public String getFilterByFirstName() {
        return filterByFirstName;
    }

    public void setFilterByFirstName(String filterByFirstName) {
        this.filterByFirstName = filterByFirstName;
    }

    public String getFilterByEmail() {
        return filterByEmail;
    }

    public void setFilterByEmail(String filterByEmail) {
        this.filterByEmail = filterByEmail;
    }

    public String getFilterByGlobusUsername() {
        return filterByGlobusUsername;
    }

    public void setFilterByGlobusUsername(String filterByGlobusUsername) {
        this.filterByGlobusUsername = filterByGlobusUsername;
    }

    public String getFilterByDescription() {
        return filterByDescription;
    }

    public void setFilterByDescription(String filterByDescription) {
        this.filterByDescription = filterByDescription;
    }

    public String getSelectFilterByBadge() {
        return selectFilterByBadge;
    }

    public void setSelectFilterByBadge(String selectFilterByBadge) {
        this.selectFilterByBadge = selectFilterByBadge;
    }

    public String getSelectFilterByUsername() {
        return selectFilterByUsername;
    }

    public void setSelectFilterByUsername(String selectFilterByUsername) {
        this.selectFilterByUsername = selectFilterByUsername;
    }

    public String getSelectFilterByLastName() {
        return selectFilterByLastName;
    }

    public void setSelectFilterByLastName(String selectFilterByLastName) {
        this.selectFilterByLastName = selectFilterByLastName;
    }

    public String getSelectFilterByFirstName() {
        return selectFilterByFirstName;
    }

    public void setSelectFilterByFirstName(String selectFilterByFirstName) {
        this.selectFilterByFirstName = selectFilterByFirstName;
    }

    public String getSelectFilterByEmail() {
        return selectFilterByEmail;
    }

    public void setSelectFilterByEmail(String selectFilterByEmail) {
        this.selectFilterByEmail = selectFilterByEmail;
    }

    public String getSelectFilterByGlobusUsername() {
        return selectFilterByGlobusUsername;
    }

    public void setSelectFilterByGlobusUsername(String selectFilterByGlobusUsername) {
        this.selectFilterByGlobusUsername = selectFilterByGlobusUsername;
    }

    public String getSelectFilterByDescription() {
        return selectFilterByDescription;
    }

    public void setSelectFilterByDescription(String selectFilterByDescription) {
        this.selectFilterByDescription = selectFilterByDescription;
    }

    @FacesConverter(forClass = UserInfo.class)
    public static class UserInfoControllerConverter implements Converter {

        @Override
        public Object getAsObject(FacesContext facesContext, UIComponent component, String value) {
            if (value == null || value.length() == 0) {
                return null;
            }
            UserInfoController controller = (UserInfoController) facesContext.getApplication().getELResolver().
                    getValue(facesContext.getELContext(), null, "userInfoController");
            return controller.getEntity(getKey(value));
        }

        java.lang.Integer getKey(String value) {
            java.lang.Integer key;
            key = Integer.valueOf(value);
            return key;
        }

        String getStringKey(java.lang.Integer value) {
            StringBuilder sb = new StringBuilder();
            sb.append(value);
            return sb.toString();
        }

        @Override
        public String getAsString(FacesContext facesContext, UIComponent component, Object object) {
            if (object == null) {
                return null;
            }
            if (object instanceof UserInfo) {
                UserInfo o = (UserInfo) object;
                return getStringKey(o.getId());
            } else {
                throw new IllegalArgumentException("Object " + object + " is of type " + object.getClass().getName() + "; expected type: " + UserInfo.class.getName());
            }
        }

    }

}
